import axios from "axios";
import type { AxiosInstance, AxiosBasicCredentials } from "axios";
import type { WorkflowModeOut, WorkflowModeIn } from "@/client";

export function requiredDocumentationFiles(
  modes?: WorkflowModeIn[] | WorkflowModeOut[],
): string[] {
  const list = ["CHANGELOG.md", "README.md", "docs/usage.md", "docs/output.md"];
  if (modes && modes.length > 0) {
    list.push(...modes.map((mode) => mode.schema_path));
  } else {
    list.push("nextflow_schema.json");
  }
  return list;
}

export function determineGitIcon(repo_url?: string): string {
  let gitProvider = "git-alt";
  if (repo_url != null) {
    if (repo_url.includes("github")) {
      gitProvider = "github";
    } else if (repo_url.includes("gitlab")) {
      gitProvider = "gitlab";
    } else if (repo_url.includes("bitbucket")) {
      gitProvider = "bitbucket";
    }
  }
  return "fa-brands fa-".concat(gitProvider);
}

const capturingMainScriptRegex =
  /manifest\s+{[\s\S]+mainScript\s+=\s+["'](?<scriptname>\S+)["']/;

export abstract class GitRepository {
  protected repo: URL;
  protected gitCommitHash: string;
  protected token?: string;
  protected httpClient: AxiosInstance;
  protected abstract readonly repoAvailabilityUrl: string;

  protected constructor(
    repoUrl: string,
    gitCommitHash: string,
    token?: string,
  ) {
    this.repo = new URL(repoUrl);
    this.gitCommitHash = encodeURIComponent(gitCommitHash);
    if (token != undefined && token.length > 0) {
      this.token = token;
    }
    this.httpClient = axios.create({
      maxRedirects: 2,
    });
  }

  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  async downloadFile(filepath: string): Promise<any> {
    return await this.httpClient.get(await this.downloadFileUrl(filepath));
  }

  async checkRepoAvailability(): Promise<boolean> {
    try {
      await this.httpClient.get(this.repoAvailabilityUrl);
      return true;
    } catch (e) {
      return false;
    }
  }

  async validateRepo(
    modes: WorkflowModeIn[] | WorkflowModeOut[],
  ): Promise<string[]> {
    const files = await this.getFileList();
    const missingFiles: string[] = [];
    const filteredFiles = files.filter(
      // files in base directory and docs directory
      (file) => !file.includes("/") || file.startsWith("docs/"),
    );
    // Check if main.nf is there and if not, check for an alternative name in the nextflow.config
    if (filteredFiles.findIndex((file) => file === "main.nf") === -1) {
      try {
        const nextflowConfig = (await this.downloadFile("nextflow.config"))
          .data;
        const entryScriptName = nextflowConfig.match(capturingMainScriptRegex)
          .groups?.["scriptname"];
        if (entryScriptName == undefined) {
          missingFiles.push("main.nf");
        } else if (files.findIndex((file) => file === entryScriptName) === -1) {
          missingFiles.push(entryScriptName);
        }
      } catch (e) {
        console.error(e);
        missingFiles.push("main.nf");
      }
    }

    // check mandatory files
    const requiredFileList = [
      "CHANGELOG.md",
      "README.md",
      "docs/usage.md",
      "docs/output.md",
    ];
    for (const requiredFile of requiredFileList) {
      if (filteredFiles.findIndex((file) => file === requiredFile) === -1) {
        missingFiles.push(requiredFile);
      }
    }
    // Check required files for modes
    if (modes && modes.length > 0) {
      for (const requiredFile of modes.map((mode) => mode.schema_path)) {
        if (files.findIndex((file) => file === requiredFile) === -1) {
          missingFiles.push(requiredFile);
        }
      }
    } else {
      // if there are no modes, check standard path for parameter schema
      if (
        filteredFiles.findIndex((file) => file === "nextflow_schema.json") ===
        -1
      ) {
        missingFiles.push("nextflow_schema.json");
      }
    }
    return missingFiles;
  }

  protected abstract downloadFileUrl(filepath: string): Promise<string>;

  abstract getFileList(): Promise<string[]>;

  static buildRepository(
    repoUrl: string,
    gitCommitHash: string,
    token?: string,
  ): GitRepository {
    if (repoUrl.includes("github")) {
      return new GithubRepository(repoUrl, gitCommitHash, token);
    } else if (repoUrl.includes("gitlab")) {
      return new GitlabRepository(repoUrl, gitCommitHash, token);
    }
    throw new Error("Repository is not supported.");
  }
}

class GithubRepository extends GitRepository {
  private readonly account: string;
  private readonly repoName: string;
  protected readonly repoAvailabilityUrl: string;

  constructor(repoUrl: string, gitCommitHash: string, token?: string) {
    super(repoUrl, gitCommitHash, token);
    const pathParts = this.repo.pathname.slice(1).split("/");
    this.account = pathParts[0];
    this.repoName = pathParts[1];
    this.repoAvailabilityUrl = `https://api.github.com/repos/${this.account}/${this.repoName}/git/commits/${this.gitCommitHash}`;
    if (token) {
      this.httpClient.interceptors.request.use((req) => {
        if (!req.url?.includes("raw")) {
          req.auth = {
            password: this.token,
            username: this.account,
          } as AxiosBasicCredentials;
        }
        return req;
      });
    }
    this.httpClient.interceptors.request.use((req) => {
      if (!req.url?.includes("raw")) {
        req.headers["X-GitHub-Api-Version"] = "2022-11-28";
        req.headers.setAccept("application/vnd.github.object+json");
      }
      return req;
    });
  }

  fileUrl(filepath: string): string {
    return `https://api.github.com/repos/${this.account}/${
      this.repoName
    }/contents/${encodeURIComponent(filepath)}?ref=${this.gitCommitHash}`;
  }

  protected async downloadFileUrl(filepath: string): Promise<string> {
    if (this.token == undefined) {
      return Promise.resolve(
        `https://raw.githubusercontent.com/${this.account}/${this.repoName}/${this.gitCommitHash}/${filepath}`,
      );
    }
    return (
      await this.httpClient.get(
        this.fileUrl(filepath) + `&time=${new Date().getTime()}`,
      )
    ).data["download_url"];
  }

  async getFileList(): Promise<string[]> {
    const response = await this.httpClient.get(
      `https://api.github.com/repos/${this.account}/${
        this.repoName
      }/git/trees/${this.gitCommitHash}?recursive=true`,
    );
    return response.data["tree"].map(
      (file: Record<string, string>): string => file["path"],
    );
  }
}

class GitlabRepository extends GitRepository {
  private readonly host: string;
  private readonly project: string;
  protected readonly repoAvailabilityUrl: string;

  constructor(repoUrl: string, gitCommitHash: string, token?: string) {
    super(repoUrl, gitCommitHash, token);
    const url = new URL(repoUrl);
    this.host = url.host;
    this.project = encodeURIComponent(url.pathname.slice(1));
    this.repoAvailabilityUrl = `https://${this.host}/api/v4/projects/${this.project}/repository/commits/${this.gitCommitHash}?stats=false`;
    if (token != undefined) {
      this.httpClient.interceptors.request.use((req) => {
        req.headers.setAuthorization(`Bearer ${token}`);
        return req;
      });
    }
  }

  protected downloadFileUrl(filepath: string): Promise<string> {
    return Promise.resolve(
      `https://${this.host}/api/v4/projects/${
        this.project
      }/repository/files/${encodeURIComponent(
        filepath,
      )}/raw?ref=${this.gitCommitHash}`,
    );
  }

  async getFileList(): Promise<string[]> {
    const fileList: string[] = [];
    try {
      let response = await this.httpClient.get(
        `https://${this.host}/api/v4/projects/${
          this.project
        }/repository/tree?ref=${this.gitCommitHash}&recursive=true&per_page=100&pagination=keyset&order_by=id&sort=asc`,
      );
      fileList.push(
        ...response.data.map(
          (file: Record<string, string>): string => file["path"],
        ),
      );
      while (response.headers?.["link"] != undefined) {
        const linkHeader: string = response.headers?.["link"];

        response = await this.httpClient.get(
          linkHeader.split(";")[0].slice(1, -1),
        );
        fileList.push(
          ...response.data.map(
            (file: Record<string, string>): string => file["path"],
          ),
        );
      }
      return fileList;
    } catch (e) {
      console.error(e);
      throw new Error("repo not reachable");
    }
  }
}
